import struct as _struct

from .connection import BadMessageError

class ComplexData:
    def __init__(self, type):
        self.type = type

    def getType(self):
        return self.type

class ArrayData(ComplexData, list):
    def __init__(self, type):
        ComplexData.__init__(self, type)
        list.__init__(self)

class NetworkTableEntryType:
    """A class defining the types supported by NetworkTables as well as
    support for serialization of those types to and from DataStreams
    """

    def __init__(self, id, name):
        self.id = id
        self.name = name

    def __str__(self):
        return "NetworkTable type: %s" % self.name

    def writeBytes(self, b, value):
        """Append a sequence of bytes to the array suitable for writing to a stream
        :param b: bytearray to append bytes to
        :param value: the value to send
        """
        raise NotImplementedError

    def readValue(self, rstream):
        """read a value from a data input stream
        :param rstream: the stream to read a value from
        :returns: the value read from the stream
        """
        raise NotImplementedError

class BasicEntryType(NetworkTableEntryType):
    def __init__(self, id, name, STRUCT):
        NetworkTableEntryType.__init__(self, id, name)
        self.STRUCT = _struct.Struct(STRUCT)

    def writeBytes(self, b, value):
        b.extend(self.STRUCT.pack(value))

    def readValue(self, rstream):
        return rstream.readStruct(self.STRUCT)[0]

class StringEntryType(NetworkTableEntryType):
    """a string type
    """
    LEN = _struct.Struct('>H')

    def __init__(self, id, name):
        NetworkTableEntryType.__init__(self, id, name)

    def writeBytes(self, b, value):
        s = value.encode('utf-8')
        b.extend(self.LEN.pack(len(s)))
        b.extend(s)

    def readValue(self, rstream):
        sLen = rstream.readStruct(self.LEN)[0]
        try:
            return rstream.read(sLen).decode('utf-8')
        except UnicodeDecodeError as e:
            raise BadMessageError(e)

class DefaultEntryTypes:
    BOOLEAN_RAW_ID = 0x00
    DOUBLE_RAW_ID = 0x01
    STRING_RAW_ID = 0x02

    BOOLEAN = BasicEntryType(BOOLEAN_RAW_ID, "Boolean", '?')
    DOUBLE = BasicEntryType(DOUBLE_RAW_ID, "Double", '>d')
    STRING = StringEntryType(STRING_RAW_ID, "String")

class ComplexEntryType(NetworkTableEntryType):
    def __init__(self, id, name):
        NetworkTableEntryType.__init__(self, id, name)

    def internalizeValue(self, key, externalRepresentation, currentInteralValue):
        raise NotImplementedError

    def exportValue(self, key, internalData, externalRepresentation):
        raise NotImplementedError

class ArrayEntryType(ComplexEntryType):
    
    LEN = _struct.Struct('>B')
    
    #TODO allow for array of complex type
    def __init__(self, id, elementType, externalArrayType):
        ComplexEntryType.__init__(self, id, "Array of [%s]" % elementType.name)
        if not issubclass(externalArrayType, ArrayData):
            raise TypeError("External Array Data Type must extend ArrayData")
        self.externalArrayType = externalArrayType
        self.elementType = elementType

    def writeBytes(self, b, value):
        if len(value) > 255:
            raise IOError("Cannot write %s as %s. Arrays have a max length of 255 values" % (value, self.name))
        b.extend(self.LEN.pack(len(value)))
        for v in value:
            self.elementType.writeBytes(b, v)

    def readValue(self, rstream):
        sLen = rstream.readStruct(self.LEN)[0]
        dataArray = [] #TODO cache object arrays
        for _ in range(sLen):
            dataArray.append(self.elementType.readValue(rstream))
        return dataArray

    def internalizeValue(self, key, externalRepresentation, currentInternalValue):
        if not isinstance(externalRepresentation, self.externalArrayType):
            raise TypeError("%s is not a %s" % (externalRepresentation, self.externalArrayType))

        if currentInternalValue is not None and \
           len(currentInternalValue) == len(externalRepresentation):
            internalArray = currentInternalValue
            del internalArray[:]
        else:
            internalArray = []
        internalArray.extend(externalRepresentation)
        return internalArray

    def exportValue(self, key, internalData, externalRepresentation):
        if not isinstance(externalRepresentation, self.externalArrayType):
            raise TypeError("%s is not a %s" % (externalRepresentation, self.externalArrayType))

        del externalRepresentation[:]
        externalRepresentation.extend(internalData)

class BooleanArray(ArrayData):
    '''
        A list of boolean values, can be used like a python list. Use
        ``putValue`` and ``retrieveValue`` to get/store this on a particular
        key in a :class:`.NetworkTable`.
    '''
    
    BOOLEAN_ARRAY_RAW_ID = 0x10
    
    @staticmethod
    def from_list(iterable):
        '''Convenience method to construct a BooleanArray from a list'''
        a = BooleanArray()
        a.extend(iterable)
        return a

    def __init__(self):
        ArrayData.__init__(self, BooleanArray.TYPE)

    def __contains__(self, key):
        return ArrayData.__contains__(self, bool(key))

    def __setitem__(self, key, value):
        ArrayData.__setitem__(self, key, bool(value))

    def append(self, obj):
        ArrayData.append(self, bool(obj))

    def extend(self, iterable):
        ArrayData.extend(self, (bool(x) for x in iterable))

    def insert(self, index, obj):
        ArrayData.insert(self, index, bool(obj))

    def remove(self, value):
        ArrayData.remove(self, bool(value))

BooleanArray.TYPE = ArrayEntryType(BooleanArray.BOOLEAN_ARRAY_RAW_ID,
                                   DefaultEntryTypes.BOOLEAN,
                                   BooleanArray)

class NumberArray(ArrayData):
    '''
        A list of number values, can be used like a python list. Use
        ``putValue`` and ``retrieveValue`` to get/store this on a particular
        key in a :class:`.NetworkTable`.
    '''
    
    NUMBER_ARRAY_RAW_ID = 0x11
    
    @staticmethod
    def from_list(iterable):
        '''Convenience method to construct a NumberArray from a list'''
        a = NumberArray()
        a.extend(iterable)
        return a

    def __init__(self):
        ArrayData.__init__(self, NumberArray.TYPE)

    def __contains__(self, key):
        return ArrayData.__contains__(self, float(key))

    def __setitem__(self, key, value):
        ArrayData.__setitem__(self, key, float(value))

    def append(self, obj):
        ArrayData.append(self, float(obj))

    def extend(self, iterable):
        ArrayData.extend(self, (float(x) for x in iterable))

    def insert(self, index, obj):
        ArrayData.insert(self, index, float(obj))

    def remove(self, value):
        ArrayData.remove(self, float(value))

NumberArray.TYPE = ArrayEntryType(NumberArray.NUMBER_ARRAY_RAW_ID,
                                  DefaultEntryTypes.DOUBLE,
                                  NumberArray)

class StringArray(ArrayData):
    '''
        A list of string values, can be used like a python list. Use
        ``putValue`` and ``retrieveValue`` to get/store this on a particular
        key in a :class:`.NetworkTable`.
    '''
    
    STRING_ARRAY_RAW_ID = 0x12
    
    @staticmethod
    def from_list(iterable):
        '''Convenience method to construct a StringArray from a list'''
        a = StringArray()
        a.extend(iterable)
        return a

    def __init__(self):
        ArrayData.__init__(self, StringArray.TYPE)

    def __contains__(self, key):
        return ArrayData.__contains__(self, str(key))

    def __setitem__(self, key, value):
        ArrayData.__setitem__(self, key, str(value))

    def append(self, obj):
        ArrayData.append(self, str(obj))

    def extend(self, iterable):
        ArrayData.extend(self, (str(x) for x in iterable))

    def insert(self, index, obj):
        ArrayData.insert(self, index, str(obj))

    def remove(self, value):
        ArrayData.remove(self, str(value))

StringArray.TYPE = ArrayEntryType(StringArray.STRING_ARRAY_RAW_ID,
                                  DefaultEntryTypes.STRING,
                                  StringArray)

class NetworkTableEntryTypeManager:
    def __init__(self):
        self.typeMap = {}
        self.registerType(DefaultEntryTypes.BOOLEAN)
        self.registerType(DefaultEntryTypes.DOUBLE)
        self.registerType(DefaultEntryTypes.STRING)
        self.registerType(BooleanArray.TYPE)
        self.registerType(NumberArray.TYPE)
        self.registerType(StringArray.TYPE)

    def getType(self, id):
        return self.typeMap.get(id)

    def registerType(self, type):
        self.typeMap[type.id] = type




import struct

__all__ = ["BadMessageError", "PROTOCOL_REVISION",
           "KEEP_ALIVE", "CLIENT_HELLO", "PROTOCOL_UNSUPPORTED",
           "SERVER_HELLO_COMPLETE", "ENTRY_ASSIGNMENT", "FIELD_UPDATE"]

# The definitions of all of the protocol message types

class BadMessageError(IOError):
    pass

PROTOCOL_REVISION = 0x0200

class Message:
    def __init__(self, HEADER, STRUCT=None):
        self.HEADER = HEADER
        if STRUCT is None:
            self.STRUCT = None
        else:
            self.STRUCT = struct.Struct(STRUCT)

    def getBytes(self, *args):
        b = bytearray(self.HEADER)
        if self.STRUCT is not None:
            b.extend(self.STRUCT.pack(*args))
        return b
    
    def read(self, rstream):
        return rstream.readStruct(self.STRUCT)

class NamedMessage(Message):
    NAME_LEN_STRUCT = struct.Struct('>H')

    def __init__(self, HEADER, STRUCT=None):
        Message.__init__(self, HEADER, STRUCT)
        
    def getBytes(self, name, *args):
        b = bytearray(self.HEADER)
        name = name.encode('utf-8')
        b.extend(self.NAME_LEN_STRUCT.pack(len(name)))
        b.extend(name)
        if self.STRUCT is not None:
            b.extend(self.STRUCT.pack(*args))
        return b

    def read(self, rstream):
        nameLen = rstream.readStruct(self.NAME_LEN_STRUCT)[0]
        try:
            name = rstream.read(nameLen).decode('utf-8')
        except UnicodeDecodeError as e:
            raise BadMessageError(e)
        return name, rstream.readStruct(self.STRUCT)

# A keep alive message that the client sends
KEEP_ALIVE = Message(b'\x00')
# A client hello message that a client sends
CLIENT_HELLO = Message(b'\x01', '>H')
# A protocol version unsupported message that the server sends to a client
PROTOCOL_UNSUPPORTED = Message(b'\x02', '>H')
# A server hello complete message that a server sends
SERVER_HELLO_COMPLETE = Message(b'\x03')
# An entry assignment message
ENTRY_ASSIGNMENT = NamedMessage(b'\x10', '>bHH')
# A field update message
FIELD_UPDATE = Message(b'\x11', '>HH')




from . import _impl
from .common import *
from .connection import *
from .entry import NetworkTableEntry
from .networktablenode import NetworkTableNode
from .type import NetworkTableEntryTypeManager


import logging
logger = logging.getLogger('nt')

__all__ = ["NetworkTableClient"]

class ClientConnectionState:
    """Represents a state that the client is in
    """

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

# indicates that the client is disconnected from the server
DISCONNECTED_FROM_SERVER = ClientConnectionState("DISCONNECTED_FROM_SERVER")

# indicates that the client is connected to the server but has not yet begun
# communication
CONNECTED_TO_SERVER = ClientConnectionState("CONNECTED_TO_SERVER")

# represents that the client has sent the hello to the server and is waiting
# for a response
SENT_HELLO_TO_SERVER = ClientConnectionState("SENT_HELLO_TO_SERVER")

# represents that the client is now in sync with the server
IN_SYNC_WITH_SERVER = ClientConnectionState("IN_SYNC_WITH_SERVER")

class ProtocolUnsupportedByServer(ClientConnectionState):
    """Represents that a client received a message from the server indicating
    that the client's protocol revision is not supported by the server
    """

    def __init__(self, serverVersion):
        """Create a new protocol unsupported state
        :param serverVersion:
        """
        ClientConnectionState.__init__(self, "PROTOCOL_UNSUPPORTED_BY_SERVER")
        self.serverVersion = serverVersion

    def getServerVersion(self):
        """:returns: the protocol version that the server reported it supports
        """
        return self.serverVersion

    def __str__(self):
        return "PROTOCOL_UNSUPPORTED_BY_SERVER: Server Version: 0x%x" % self.serverVersion

class ClientError(ClientConnectionState):
    """Represents that the client is in an error state
    """

    def __init__(self, e):
        """Create a new error state
        :param e:
        """
        ClientConnectionState.__init__(self, "CLIENT_ERROR")
        self.e = e

    def getException(self):
        """:returns: the exception that caused the client to enter an
            error state
        """
        return self.e

    def __str__(self):
        return "CLIENT_ERROR: %s" % self.e

class ClientConnectionAdapter:
    """Object that adapts messages from a server
    
    There should only be one instance of this object
    """

    def gotoState(self, newState):
        # Always must hold the lock when calling this
        if self.connectionState != newState:
            logger.info("%s entered connection state: %s", self, newState)
            if newState == IN_SYNC_WITH_SERVER:
                self.connectionListenerManager.fireConnectedEvent()
            if self.connectionState == IN_SYNC_WITH_SERVER:
                self.connectionListenerManager.fireDisconnectedEvent()
            self.connectionState = newState

    def getConnectionState(self):
        """:returns: the state of the connection
        """
        return self.connectionState

    def isConnected(self):
        """:returns: if the client is connected to the server
        """
        return self.connectionState == IN_SYNC_WITH_SERVER

    def __init__(self, entryStore, streamFactory, connectionListenerManager, typeManager):
        """Create a new ClientConnectionAdapter
        :param entryStore:
        :param streamFactory:
        :param transactionPool:
        :param connectionListenerManager:
        """
        self.entryStore = entryStore
        self.streamFactory = streamFactory
        self.connectionListenerManager = connectionListenerManager
        self.typeManager = typeManager
        self.connection = None
        self.readManager = None
        self.connectionState = DISCONNECTED_FROM_SERVER
        self.connectionLock = _impl.create_rlock('client_conn_lock')
        
    def __str__(self):
        return 'Client 0x%08x' % id(self)
    
    #
    # Errors
    #
    
    def badMessage(self, e):
        self.close(ClientError(e))

    def ioError(self, e):
        if self.connectionState != DISCONNECTED_FROM_SERVER:
            # will get io exception when on read thread connection is closed
            self.reconnect()
        #self.gotoState(ClientError(e))
    

    #
    # Connection management
    #

    def reconnect(self):
        """Reconnect the client to the server (even if the client is not
        currently connected)
        """
        
        with self.connectionLock:
            self.close() #close the existing stream and monitor thread if needed
            try:
                stream = self.streamFactory.createStream()
                if stream is None:
                    return
                self.connection = NetworkTableConnection(stream, self.typeManager)
                self.readManager = ReadManager(self,
                        self.connection, name="Client Connection Reader Thread")
                self.readManager.start()
                self.connection.sendClientHello()
                self.gotoState(CONNECTED_TO_SERVER)
            except IOError:
                self.close() #make sure to clean everything up if we fail to connect

    def close(self, newState=DISCONNECTED_FROM_SERVER):
        """Close the connection to the server and enter the given state
        :param newState: new state; defaults to DISCONNECTED_FROM_SERVER.
        """
        
        with self.connectionLock:
            self.gotoState(newState)
            if self.readManager is not None:
                self.readManager.stop()
                self.readManager = None
            if self.connection is not None:
                self.connection.close()
                self.connection = None
            self.entryStore.clearIds()

    def getEntry(self, id):
        return self.entryStore.getEntry(id)

    def keepAlive(self):
        pass

    def clientHello(self, protocolRevision):
        raise BadMessageError("A client should not receive a client hello message")

    def protocolVersionUnsupported(self, protocolRevision):
        with self.connectionLock:
            self.close()
            self.gotoState(ProtocolUnsupportedByServer(protocolRevision))

    def serverHelloComplete(self):
        with self.connectionLock:
            if self.connectionState == CONNECTED_TO_SERVER:
                try:
                    self.gotoState(IN_SYNC_WITH_SERVER)
                    self.entryStore.sendUnknownEntries(self.connection)
                except IOError as e:
                    self.ioError(e)
            else:
                raise BadMessageError("A client should only receive a server hello " +
                                      "complete once and only after it has connected " +
                                      "to the server (state is %s)" % self.connectionState)

    def offerIncomingAssignment(self, entry):
        self.entryStore.offerIncomingAssignment(entry)

    def offerIncomingUpdate(self, entry, sequenceNumber, value):
        self.entryStore.offerIncomingUpdate(entry, sequenceNumber, value)

    def sendEntry(self, entryBytes):
        try:
            with self.connectionLock:
                if self.connectionState == IN_SYNC_WITH_SERVER:
                    self.connection.sendEntry(entryBytes)
        except IOError as e:
            self.ioError(e)
    
    def flush(self):
        with self.connectionLock:
            if self.connection is not None:
                try:
                    self.connection.flush()
                except IOError as e:
                    self.ioError(e)

    def ensureAlive(self):
        with self.connectionLock:
            if self.connection is not None:
                try:
                    self.connection.sendKeepAlive()
                except IOError as e:
                    self.ioError(e)
            else:
                self.reconnect() #try to reconnect if not connected

class ClientNetworkTableEntryStore(AbstractNetworkTableEntryStore):
    """The entry store for a NetworkTableClient
    """

    def addEntry(self, newEntry):
        with self.entry_lock:
            entry = self.namedEntries.get(newEntry.name)

            if entry is not None:
                if entry.getId() != newEntry.getId():
                    self.idEntries.pop(entry.getId(), None)
                    if newEntry.getId() != NetworkTableEntry.UNKNOWN_ID:
                        entry.setId(newEntry.getId())
                        self.idEntries[newEntry.getId()] = entry

                entry.forcePut(newEntry.getSequenceNumber(),
                               newEntry.getValue(), type=newEntry.getType())
            else:
                if newEntry.getId() != NetworkTableEntry.UNKNOWN_ID:
                    self.idEntries[newEntry.getId()] = newEntry
                self.namedEntries[newEntry.name] = newEntry
        return True

    def updateEntry(self, entry, sequenceNumber, value):
        with self.entry_lock:
            entry.forcePut(sequenceNumber, value)
            if entry.getId() == NetworkTableEntry.UNKNOWN_ID:
                return False
            return True

    def sendUnknownEntries(self, connection):
        """Send all unknown entries in the entry store to the given connection
        :param connection:
        """
        transaction = []
        with self.entry_lock:
            # Cannot hold the entry lock when calling sendEntry
            for entry in self.namedEntries.values():
                if entry.getId() == NetworkTableEntry.UNKNOWN_ID:
                    transaction.append(entry.getAssignmentBytes())
                    
        for entry in transaction:
            connection.sendEntry(entry)
        connection.flush()

class NetworkTableClient(NetworkTableNode):
    """A client node in NetworkTables 2.0
    
    There should only be one instance of this object
    """

    def __init__(self, streamFactory):
        """Create a new NetworkTable Client
        :param streamFactory:
        """
        NetworkTableNode.__init__(self, ClientNetworkTableEntryStore(self))
        typeManager = NetworkTableEntryTypeManager()
        self.adapter = ClientConnectionAdapter(self.entryStore, streamFactory,
                                               self, typeManager)
        self.writeManager = WriteManager(self.adapter, self.entryStore, 1.0)

        self.entryStore.setOutgoingReceiver(self.writeManager)
        self.entryStore.setIncomingReceiver(None)
        self.writeManager.start()

    def reconnect(self):
        """force the client to disconnect and reconnect to the server again.
        Will connect if the client is currently disconnected
        """
        self.adapter.reconnect()

    def close(self):
        self.adapter.close()

    def stop(self):
        self.writeManager.stop()
        self.close()

    def isConnected(self):
        return self.adapter.isConnected()

    def isServer(self):
        return False
